#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test shows a surrogate-assisted Bayesian calibration of a time dependent
    analytical function.

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

Created on Fri Aug 9 2019

"""

import numpy as np
import pandas as pd
import sys
import joblib
# import bayesvalidrox
# Add BayesValidRox path
#sys.path.append("../../src/bayesvalidrox/")

import bayesvalidrox as bv
from bayesvalidrox.pylink.pylink import PyLinkForwardModel
from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.surrogate_models import MetaModel
from bayesvalidrox.surrogate_models.meta_model_engine import MetaModelEngine
from bayesvalidrox.post_processing.post_processing import PostProcessing
from bayesvalidrox.bayes_inference.bayes_inference import BayesInference
from bayesvalidrox.bayes_inference.discrepancy import Discrepancy
import matplotlib
matplotlib.use('agg')


if __name__ == "__main__":

    # Number of parameters
    ndim = 10  # 2, 10

    # =====================================================
    # =============   COMPUTATIONAL MODEL  ================
    # =====================================================
    Model = PyLinkForwardModel()

    Model.link_type = 'Function'
    Model.py_file = 'analytical_function'
    Model.name = 'AnalyticFunc'

    Model.Output.names = ['Z']

    # For Bayesian inversion synthetic data with X=[0,0]
    Model.observations = {}
    Model.observations['Time [s]'] = np.arange(0, 10, 1.) / 9
    Model.observations['Z'] = np.repeat([2.], 10)

    # For Checking with the MonteCarlo refrence
    Model.mc_reference = {}
    Model.mc_reference['Time [s]'] = np.arange(0, 10, 1.) / 9
    Model.mc_reference['mean'] = np.load(f"data/mean_{ndim}.npy")
    Model.mc_reference['std'] = np.load(f"data/std_{ndim}.npy")

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define the uncertain parameters with their mean and
    # standard deviation
    Inputs = Input()

    # Assuming dependent input variables
    # Inputs.Rosenblatt = True

    for i in range(ndim):
        Inputs.add_marginals()
        Inputs.Marginals[i].name = "$\\theta_{"+str(i+1)+"}$"
        Inputs.Marginals[i].dist_type = 'uniform'
        Inputs.Marginals[i].parameters = [-5, 5]

    # arbitrary polynomial chaos
    # inputParams = np.load('data/InputParameters_{}.npy'.format(ndim))
    # for i in range(ndim):
    #     Inputs.add_marginals()
    #     Inputs.Marginals[i].name = f'$X_{i+1}$'
    #     Inputs.Marginals[i].input_data = inputParams[:, i]

    # =====================================================
    # ==========  DEFINITION OF THE METAMODEL  ============
    # =====================================================
    MetaModelOpts = MetaModel(Inputs, Model)

    # Select if you want to preserve the spatial/temporal depencencies
    # MetaModelOpts.dim_red_method = 'PCA'
    # MetaModelOpts.var_pca_threshold = 99.999
    # MetaModelOpts.n_pca_components = 10

    # Select your metamodel method
    # 1) PCE (Polynomial Chaos Expansion) 2) aPCE (arbitrary PCE)
    # 3) GPE (Gaussian Process Emulator)
    MetaModelOpts.meta_model_type = 'aPCE'

    # ------------------------------------------------
    # ------------- PCE Specification ----------------
    # ------------------------------------------------
    # Select the sparse least-square minimization method for
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)FastARD: Fast Bayesian ARD Regression
    # 6)BCS: Bayesian Compressive Sensing
    # 7)OMP: Orthogonal Matching Pursuit
    # 8)VBL: Variational Bayesian Learning
    # 9)EBL: Emperical Bayesian Learning
    MetaModelOpts.pce_reg_method = 'FastARD'

    # Bootstraping
    # 1) normal 2) fast
    MetaModelOpts.bootstrap_method = 'fast'
    MetaModelOpts.n_bootstrap_itrs = 1

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final
    # metamodel. pce_deg accepts degree as a scalar or a range.
    MetaModelOpts.pce_deg = 12

    # q-quasi-norm 0<q<1 (default=1)
    MetaModelOpts.pce_q_norm = 0.85 if ndim < 5 else 0.5

    # Print summary of the regression results
    # MetaModelOpts.verbose = True

    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    MetaModelOpts.add_ExpDesign()

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    MetaModelOpts.ExpDesign.method = 'sequential'
    MetaModelOpts.ExpDesign.n_init_samples = 3*ndim

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley
    # 6) chebyshev(FT) 7) grid(FT) 8)user
    MetaModelOpts.ExpDesign.sampling_method = 'latin_hypercube'

    # Provide the experimental design object with a hdf5 file
    # MetaModelOpts.ExpDesign.hdf5_file = 'ExpDesign_AnalyticFunc.hdf5'

    # ------------------------------------------------
    # ------- Sequential Design configuration --------
    # ------------------------------------------------
    # Set the sampling parameters
    MetaModelOpts.ExpDesign.n_new_samples = 1
    MetaModelOpts.ExpDesign.n_max_samples = 10#150
    MetaModelOpts.ExpDesign.mod_LOO_threshold = 1e-16

    # MetaModelOpts.adapt_verbose = True
    # 1) None 2) 'equal' 3)'epsilon-decreasing' 4) 'adaptive'
    MetaModelOpts.ExpDesign.tradeoff_scheme = None
    # MetaModelOpts.ExpDesign.n_replication = 5
    # -------- Exploration ------
    # 1)'Voronoi' 2)'random' 3)'latin_hypercube' 4)'LOOCV' 5)'dual annealing'
    MetaModelOpts.ExpDesign.explore_method = 'random'

    # Use when 'dual annealing' chosen
    MetaModelOpts.ExpDesign.max_func_itr = 1000

    # Use when 'Voronoi' or 'random' or 'latin_hypercube' chosen
    MetaModelOpts.ExpDesign.n_canddidate = 1000
    MetaModelOpts.ExpDesign.n_cand_groups = 4

    # -------- Exploitation ------
    # 1)'BayesOptDesign' 2)'BayesActDesign' 3)'VarOptDesign' 4)'alphabetic'
    # 5)'Space-filling'
    MetaModelOpts.ExpDesign.exploit_method = 'BayesActDesign'

    # BayesActDesign -> when data is available
    # 1) BME (Bayesian model evidence) 2) infEntropy (Information entropy)
    # 2)DKL (Kullback-Leibler Divergence)
    MetaModelOpts.ExpDesign.util_func = 'DKL'

    # Defining the measurement error, if it's known a priori
    obsData = pd.DataFrame(Model.observations, columns=Model.Output.names)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData**2
    MetaModelOpts.Discrepancy = DiscrepancyOpts

    # Plot the posterior snapshots for SeqDesign
    MetaModelOpts.ExpDesign.post_snapshot = False
    MetaModelOpts.ExpDesign.step_snapshot = 1
    MetaModelOpts.ExpDesign.max_a_post = [0] * ndim

    # For calculation of validation error for SeqDesign
    prior = np.load(f"data/Prior_{ndim}.npy")
    prior_outputs = np.load(f"data/origModelOutput_{ndim}.npy")
    likelihood = np.load(f"data/validLikelihoods_{ndim}.npy")
    MetaModelOpts.valid_samples = prior[:500]
    MetaModelOpts.valid_model_runs = {'Z': prior_outputs[:500]}
    # MetaModelOpts.valid_likelihoods = likelihood

    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # Train the meta model
    meta_model_engine = MetaModelEngine(MetaModelOpts)
    #meta_model_engine.run()
    PCEModel = meta_model_engine.MetaModel


    # =====================================================
    # ========  Bayesian inference with Emulator ==========
    # =====================================================
    BayesOpts = BayesInference(PCEModel)
    BayesOpts.emulator = False
    BayesOpts.plot_post_pred = True

    # Select the inference method
    import emcee
    BayesOpts.inference_method = "MCMC"
    # Set the MCMC parameters passed to self.mcmc_params
    BayesOpts.mcmc_params = {
        'n_steps': 1e3,
        'n_walkers': 30,
        'moves': emcee.moves.KDEMove(),
        'multiprocessing': False,
        'verbose': False
        }

    # ----- Define the discrepancy model -------
    BayesOpts.measurement_error = obsData

    # # -- (Option B) --
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData**2
    BayesOpts.Discrepancy = DiscrepancyOpts

    # Start the calibration/inference
    Bayes_PCE = BayesOpts.create_inference()

    # Save class objects
    with open(f'Bayes_{Model.name}.pkl', 'wb') as output:
        joblib.dump(Bayes_PCE, output, 2)
